local options = {}
local isInCombat = false

local DazzlingRodTournament = LibStub("AceAddon-3.0"):NewAddon("DazzlingRodTournament", "AceConsole-3.0", "AceEvent-3.0")
local AceGUI = LibStub("AceGUI-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local AceConfigCommand = LibStub("AceConfigCmd-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("DazzlingRodTournament")

local defaults = {
  profile = {
    mode = "Team vs. Team",
    friendlyfire = false,
    group_1 = 'A',
    group_2 = 'A',
    group_3 = 'A',
    group_4 = 'B',
    group_5 = 'B',
    group_6 = 'B',
    group_7 = 'Zuschauer',
    group_8 = 'Zuschauer',
    marker_leader = 1,
    marker_hottest_streak = 3,
    marker_last_hit = 8,
  }
}

local state = {
  isInRaid = false,
  isInParty = false,
  channel = "SAY",
  gameIsRunning = false,
  hottest_streak = { player = nil, streak = 0 },
  highest_score = { player = nil, score = 0 }
}

local players = {
-- [unitName] = {
--   score = 0,
--   streak = 0,
--   team = 'none'
-- }
}

local teams = {
  A = {
    players = {},
    score = 0
  },
  B = {
    players = {},
    score = 0
  }
}



function DazzlingRodTournament:COMBAT_LOG_EVENT_UNFILTERED(event, timestamp, message, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, ...)
  if ((message == "SPELL_AURA_APPLIED" or message == "SPELL_AURA_REFRESH" or message == "SPELL_AURA_APPLIED_DOSE") ) then
    spellID, spellName, moredata = ...

    if (spellID == 187973 and state.gameIsRunning) then 

      -- DEBUG:DazzlingRodTournament:Print("hit")

      if (DazzlingRodTournament:IsValidHit(sourceName, destName)) then

        -- check if it is the first hit in the game
        if state.highest_score.player == nil then 
          if state.isInRaid then
            SendChatMessage("FIRST BLOOD - " .. sourceName, "RAID_WARNING", nil, nil)
          else
            SendChatMessage("FIRST BLOOD - " .. sourceName, "PARTY", nil, nil)
          end
        end 

        -- set the raid marker for the last hit
        SetRaidTarget(players[destName].unitID, self.db.global.marker_last_hit)


        -- DEBUG:DazzlingRodTournament:Print("valid hit")
        -- add the point to the player
        players[sourceName].score = players[sourceName].score + 1

        -- check if this player is the leader for the highest score currently
        if state.highest_score.player ~= sourceName and state.highest_score.score < players[sourceName].score then
          state.highest_score.player = sourceName
          state.highest_score.score = players[sourceName].score
          SetRaidTarget(players[sourceName].unitID, self.db.global.marker_leader)
        end

        -- add the point to the streak
        players[sourceName].streak = players[sourceName].streak + 1

        -- ultrakills or something like that?
        if players[sourceName].streak == 5 then 
          if state.isInRaid then
            SendChatMessage("KILLING SPREE - " .. sourceName, "RAID_WARNING", nil, nil)
          else
            SendChatMessage("KILLING SPREE - " .. sourceName, "PARTY", nil, nil)
          end
        end 
        if players[sourceName].streak == 10 then 
          if state.isInRaid then
            SendChatMessage("RAMPAGE - " .. sourceName, "RAID_WARNING", nil, nil)
          else
            SendChatMessage("RAMPAGE - " .. sourceName, "PARTY", nil, nil)
          end
        end 
        if players[sourceName].streak == 15 then 
          if state.isInRaid then
            SendChatMessage("DOMINATING - " .. sourceName, "RAID_WARNING", nil, nil)
          else
            SendChatMessage("DOMINATING - " .. sourceName, "PARTY", nil, nil)
          end
        end 
        if players[sourceName].streak == 20 then 
          if state.isInRaid then
            SendChatMessage("UNSTOPPABLE - " .. sourceName, "RAID_WARNING", nil, nil)
          else
            SendChatMessage("UNSTOPPABLE - " .. sourceName, "PARTY", nil, nil)
          end
        end 
        if players[sourceName].streak == 25 then 
          if state.isInRaid then
            SendChatMessage("GODLIKE - " .. sourceName, "RAID_WARNING", nil, nil)
          else
            SendChatMessage("GODLIKE - " .. sourceName, "PARTY", nil, nil)
          end
        end 
        if players[sourceName].streak == 25 then 
          if state.isInRaid then
            SendChatMessage("WICKED SICK - " .. sourceName, "RAID_WARNING", nil, nil)
          else
            SendChatMessage("WICKED SICK - " .. sourceName, "PARTY", nil, nil)
          end
        end 


        -- new highest streak?
        if (players[sourceName].highest_streak < players[sourceName].streak) then

          -- check if this player is the leader for the highest streak in game
          if state.hottest_streak.player ~= sourceName and state.hottest_streak.streak < players[sourceName].streak then
            state.hottest_streak.player = sourceName
            state.hottest_streak.streak = players[sourceName].streak
            -- set the raid marker
            SetRaidTarget(players[sourceName].unitID, self.db.global.marker_hottest_streak)
          end


          players[sourceName].highest_streak = players[sourceName].streak

          -- DEBUG: DazzlingRodTournament:Print("New Highest Streak: " .. players[sourceName].highest_streak)

        end

        -- add the point to the team, if in team vs. team
        if self.db.global.mode == "Team vs. Team" then
          teams[players[sourceName].team].score = teams[players[sourceName].team].score + 1
        end

        -- is the streak of the destination cancelled?
        -- DEBUG:local cancels = DazzlingRodTournament:CancelsStreak(sourceName, destName)
        -- DEBUG:DazzlingRodTournament:Print("Cancels Streak? " .. tostring(cancels))
        if DazzlingRodTournament:CancelsStreak(sourceName, destName) then
          -- DEBUG:DazzlingRodTournament:Print("Previous Streak " .. players[destName].streak)
          players[destName].streak = 0
          -- DEBUG:DazzlingRodTournament:Print("Now Streak " .. players[destName].streak)
        end
      end
    end
  end
end


function DazzlingRodTournament:ChatCmd(input)
  if not input or input:trim() == "" then 
    AceConfigDialog:Open("Dazzling Rod Tournament")
  else
    if input:trim() == "start" then

      state.isInRaid = (GetUnitName("raid1") ~= nil)
      state.isInParty = (GetUnitName("party1") ~= nil)

      -- if the player is not in a party even, stop the game
      if (not state.isInParty and not state.isInRaid) then
        DazzlingRodTournament:Print(L["error_not_in_group"])
        return false
      end

      -- set the output channel 
      if (state.isInParty and not state.isInRaid) then
        state.channel = "PARTY"
      end
      if (state.isInRaid) then
        state.channel = "RAID"
      end

      if (state.isInParty and not state.isInRaid and self.db.global.mode == "Team vs. Team") then
        DazzlingRodTournament:Announce(L["error_team_vs_team_in_party"])
        self.db.global.mode = "Deathmatch"
      end

      DazzlingRodTournament:InitializePlayers()


    elseif input:trim() == "stop" then
      state.gameIsRunning = false
      DazzlingRodTournament:ReportResults()


    else
      AceConfigCommand.HandleCommand(DazzlingRodTournament, "Dazzling Rod Tournament", "Dazzling Rod Tournament", input:trim() ~= "help" and input or "")
    end
  end
end

function DazzlingRodTournament:Announce(message)
  SendChatMessage(message, state.channel, "Common")
end

function DazzlingRodTournament:IsValidHit(playerA, playerB) 

  -- DEBUG:DazzlingRodTournament:Announce("IsValidHit("  .. playerA ..", ".. playerB)

  -- if the source is not in team a or b, it won't count

  -- hits outside of the raid are not counted at all
  if players[playerB] == nil then 

    -- DEBUG:DazzlingRodTournament:Announce("players["..playerB.."] == nil: " .. players[playerB])
    return false 

  end

  -- if we hit a spectator, it's not a valid hit
  if (players[playerB].team == "Zuschauer" or players[playerA].team == "Zuschauer") then 
    -- DEBUG:DazzlingRodTournament:Announce("players["..playerB.."].team == Zuschauer: " .. players[playerB].team)
    return false end

  -- in deathmatch, every hit from the current party/raid is a hit
  if self.db.global.mode == "Deathmatch" then return true end

  -- in team vs. team, a hit is only valid if you hit the opposing team
  if self.db.global.mode == "Team vs. Team" then 

    
    -- DEBUG:local isInSameTeam = DazzlingRodTournament:IsInSameTeam(playerA, playerB)
   -- DEBUG: DazzlingRodTournament:Announce("in same team? " .. players[playerA].team .. "/" .. players[playerB].team .. " = " .. tostring(isInSameTeam))

    return not DazzlingRodTournament:IsInSameTeam(playerA, playerB) 
  end 
  -- else: nope!
  return false
end

function DazzlingRodTournament:CancelsStreak(playerA, playerB)
  -- hits outside of the raid are not counted at all
  if players[playerB] == nil then 
    -- DEBUG:DazzlingRodTournament:Print("playerB does not exist")
    return false 
  end
 
  -- if we hit a spectator, it's not a valid hit, so no cancel
  if (players[playerB].team == "Zuschauer") then return false end

  -- in deathmatch, every hit from the current party/raid is a hit
  if self.db.global.mode == "Deathmatch" then return true end

  -- in team vs. team, a hit is only valid if you hit the opposing team or if friendly fire is set to 'on'
  if self.db.global.mode == "Team vs. Team" then return not DazzlingRodTournament:IsInSameTeam(playerA, playerB) or self.db.global.friendlyfire end

end


function DazzlingRodTournament:IsInSameTeam(playerA, playerB)
  return players[playerA].team == players[playerB].team
end

function DazzlingRodTournament:ReportResults()
  DazzlingRodTournament:Announce(L["announce_game_ended"])
  DazzlingRodTournament:Announce("")

  if self.db.global.mode == "Team vs. Team" then

    if teams.A.score > teams.B.score then
      DazzlingRodTournament:Announce(string.format(L["announce_team_a_winner"], teams.A.score, teams.B.score))
    elseif teams.B.score > teams.A.score then  
      DazzlingRodTournament:Announce(string.format(L["announce_team_b_winner"], teams.B.score, teams.A.score))
     else
      DazzlingRodTournament:Announce(string.format(L["announce_no_winner"], teams.A.score))

    end
    DazzlingRodTournament:Announce("")
  end

    DazzlingRodTournament:Announce(L["announce_player_results"])
    local result = {}
    for p, d in pairs(players) do
      -- DEBUG:DazzlingRodTournament:Print("test")
      table.insert(result, {name = p, score = d.score, streak = d.streak, team = d.team, highest_streak = d.highest_streak })
    end

    -- sort by score
    table.sort(result, 
      function(a, b) 
        if a.score > b.score then 
          return true 
        else 
          return false 
        end
      end)

    DazzlingRodTournament:tprint(result)

    local i = 1
    for k,v in ipairs(result) do
      if self.db.global.mode == "Team vs. Team" then
        if v.team == 'A' or v.team == 'B' then 
          DazzlingRodTournament:Announce(string.format(L["announce_single_player_results_team_vs_team"], i, v.name, v.score, v.team, v.highest_streak))
        end
      else 
        if v.team == 'A' or v.team == 'B' then 
          DazzlingRodTournament:Announce(string.format(L["announce_single_player_results_deathmatch"], i, v.name, v.score))
        end
      end
      i = i + 1
    end

    -- sort by streak
    table.sort(result, 
      function(a, b) 
        if a.highest_streak > b.highest_streak then 
          return true 
        else 
          return false 
        end
      end)


    DazzlingRodTournament:Announce(L["announce_player_streaks"])
    i = 1
    for k,v in ipairs(result) do
      DazzlingRodTournament:Announce(string.format(L["announce_single_player_streak"], i, v.name, v.highest_streak))
      i = i + 1
      if i > 3 then break end
    end

    -- DEBUG:DazzlingRodTournament:tprint(result)
end


function DazzlingRodTournament:InitializePlayers()
  -- initialize the teams
  teams = {
    A = {
      players = {},
      score = 0
    },
    B = {
      players = {},
      score = 0
    }
  }

  players = {
  }

  state.hottest_streak = { player = nil, streak = 0 }
  state.highest_score = { player = nil, score = 0 }
 
  local memberCount = GetNumGroupMembers()

  if state.isInParty then
    -- if the player is in a party, GetNumGroupMembers includes the player, but theres only "player" and "party1" to "party4", so we need to adjust that loop
    memberCount = memberCount - 1
    local unitName = GetUnitName("player", true) --DazzlingRodTournament:CleanseUnitName(GetUnitName("player"))
    DazzlingRodTournament:Print("Trying unitName '".. unitName .. "'")
    players[unitName] = {
      score = 0,
      streak = 0,
      highest_streak = 0,
      team = 'none',
      unitID = 'player'
    }
  end

  local prefix = state.isInRaid and "raid" or "party"
  for i = 1, memberCount do
    local unitName = GetUnitName(prefix .. i, true) -- DazzlingRodTournament:CleanseUnitName(GetUnitName(prefix .. i))
    players[unitName] = {
      score = 0,
      streak = 0,
      highest_streak = 0,
      team = 'none',
      unitID = prefix .. i
    }

    -- set the right team
    if (state.isInRaid and self.db.global.mode == "Team vs. Team") then
      local _,_,grp = GetRaidRosterInfo(i)
      local team = self.db.global["group_" .. grp]

      -- DEBUG:DazzlingRodTournament:Announce("Group #" .. grp)

      players[unitName].team = team
      -- fill the teams table
      if (team == 'A' or team == 'B') then
        table.insert(teams[team].players, unitName)
      end
    end
  end

  DazzlingRodTournament:Announce(L["announce_game_starts"])
  DazzlingRodTournament:Announce("")
  if (self.db.global.mode == "Team vs. Team") then
    DazzlingRodTournament:Announce(string.format(L["announce_game_mode"], L["mode_team_vs_team"]))
  else 
    DazzlingRodTournament:Announce(string.format(L["announce_game_mode"], L["mode_deathmatch"]))
  end



 
  if self.db.global.mode == "Team vs. Team" then
    DazzlingRodTournament:Announce(L["team"] .. " A: " ..DazzlingRodTournament:ConcatTableToString(teams.A.players))
    DazzlingRodTournament:Announce(L["team"] .. " B: " ..DazzlingRodTournament:ConcatTableToString(teams.B.players))
  else 
    local playerNames = DazzlingRodTournament:GetTableKeys(players)
    -- DEBUG:DazzlingRodTournament:tprint(playerNames)
    DazzlingRodTournament:Announce(string.format(L["announce_players_in_deathmatch"], DazzlingRodTournament:ConcatTableToString(playerNames)))
  end

  state.gameIsRunning = true
end

function DazzlingRodTournament:ConcatTableToString(t, sep) 
  sep = sep or ", "
  local s = ""
  for k,v in pairs(t) do
    s = s .. v .. sep
  end
  return string.sub(s, 0, string.len(s) - string.len(sep) )

end

function DazzlingRodTournament:CleanseUnitName(str) 
  if string.find(str, " ") then
    return string.sub(str, 0, string.find(str, " "))
  end

  if string.find(str, "-") then
    return string.sub(str, 0, string.find(str, " "))
  end    

  return str
end


function DazzlingRodTournament:GetTableKeys(t)
  local keyset = {}
  local n = 0
  for k,v in pairs(t) do
    n = n + 1
    keyset[n] = k
  end
  return keyset
end

function DazzlingRodTournament:tprint(tbl, indent)
if not indent then indent = 3 end
  for k, v in pairs(tbl) do
    formatting = string.rep("  ", indent) .. k .. ": "
    if type(v) == "table" then
      DazzlingRodTournament:Print(formatting)
      DazzlingRodTournament:tprint(v, indent+1)
    elseif type(v) == 'boolean' then
      DazzlingRodTournament:Print(formatting .. tostring(v))      
    else
      DazzlingRodTournament:Print(formatting .. v)
    end
  end
end


function DazzlingRodTournament:OnInitialize()
 
  -- setting the database for this addon
  self.db = LibStub("AceDB-3.0"):New("DazzlingRodTournamentDB", defaults)

  local options = {
    name = "Dazzling Rod Tournament",
    handler = DazzlingRodTournament,
    type = 'group',
    args = {
      mode = {
        type = 'select',
        name = L["option_game_mode"],
        values = {["Deathmatch"]= L["mode_deathmatch"], ["Team vs. Team"] = L["mode_team_vs_team"]},
        desc = L["option_game_mode_desc"],
        set = function(info, val) self.db.global.mode = val end,
        get = function(info) return self.db.global.mode end,
        width = 'normal',
        order = 0,
      },
      friendlyfire = {
        type = 'toggle',
        name = L["option_friendly_fire"],
        desc = L["option_friendly_fire_desc"],
        set = function(info, val) self.db.global.friendlyfire = val end,
        get = function(info) return self.db.global.friendlyfire end,
        width = 'double', 
        order = 1,
      },
      marks = {
        type = 'group',
        name = L["option_marks"],
        desc = L["option_marks_desc"],
        args = {
          marker_leader = {
          type = 'select',
          name = L["option_marker_leader"],
          desc = L["option_marker_leader_desc"],
          values = {[0] = L["option_marks_none"], [1] = L["option_marks_star"], [2] = L["option_marks_circle"], [3] = L["option_marks_diamond"], [4] = L["option_marks_triagle"], [5] = L["option_marks_moon"], [6] = L["option_marks_square"], [7] = L["option_marks_cross"], [8] = L["option_marks_skull"]},
          set = function(info, val) self.db.global.marker_leader = val end,
          get = function(info) return self.db.global.marker_leader end,
          width = 'full',
          order = 2,
          },
          marker_hottest_streak = {
            type = 'select',
            name = L["option_marker_hottest_streak"],
            desc = L["option_marker_hottest_streak_desc"],
            values = {[0] = L["option_marks_none"], [1] = L["option_marks_star"], [2] = L["option_marks_circle"], [3] = L["option_marks_diamond"], [4] = L["option_marks_triagle"], [5] = L["option_marks_moon"], [6] = L["option_marks_square"], [7] = L["option_marks_cross"], [8] = L["option_marks_skull"]},
            set = function(info, val) self.db.global.marker_hottest_streak = val end,
            get = function(info) return self.db.global.marker_hottest_streak end,
            width = 'full',
            order = 3,
          },
          marker_last_hit = {
            type = 'select',
            name = L["option_marker_last_hit"],
            desc = L["option_marker_last_hit_desc"],
            values = {[0] = L["option_marks_none"], [1] = L["option_marks_star"], [2] = L["option_marks_circle"], [3] = L["option_marks_diamond"], [4] = L["option_marks_triagle"], [5] = L["option_marks_moon"], [6] = L["option_marks_square"], [7] = L["option_marks_cross"], [8] = L["option_marks_skull"]},
            set = function(info, val) self.db.global.marker_last_hit = val end,
            get = function(info) return self.db.global.marker_last_hit end,
            width = 'full',
            order = 4,
          }
        }
      },
      teams = {
        type = 'group',
        name = L["option_teams"],
        desc = L["option_teams_desc"],
        args = {
          group_1 = {
            type = 'select',
            name = L["options_choose_group_1"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_1_desc"],
            set = function(info, val) self.db.global.group_1 = val end,
            get = function(info) return self.db.global.group_1 end
          },
          group_2 = {
            type = 'select',
            name = L["options_choose_group_2"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_2_desc"],
            set = function(info, val) self.db.global.group_2 = val end,
            get = function(info) return self.db.global.group_2 end
          },
          group_3 = {
            type = 'select',
            name = L["options_choose_group_3"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_3_desc"],
            set = function(info, val) self.db.global.group_3 = val end,
            get = function(info) return self.db.global.group_3 end
          },
          group_4 = {
            type = 'select',
            name = L["options_choose_group_4"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_4_desc"],
            set = function(info, val) self.db.global.group_4 = val end,
            get = function(info) return self.db.global.group_4 end
          },
          group_5 = {
            type = 'select',
            name = L["options_choose_group_5"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_5_desc"],
            set = function(info, val) self.db.global.group_5 = val end,
            get = function(info) return self.db.global.group_5 end
          },
          group_6 = {
            type = 'select',
            name = L["options_choose_group_6"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_6_desc"],
            set = function(info, val) self.db.global.group_6 = val end,
            get = function(info) return self.db.global.group_6 end
          },
          group_7 = {
            type = 'select',
            name = L["options_choose_group_7"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_7_desc"],
            set = function(info, val) self.db.global.group_7 = val end,
            get = function(info) return self.db.global.group_7 end
          },
          group_8 = {
            type = 'select',
            name = L["options_choose_group_8"],
            values = {['A'] = L["team"] ..' A', ['B'] = L["team"] ..' B', ['Zuschauer'] = L["spectator"]},
            desc = L["options_choose_group_8_desc"],
            set = function(info, val) self.db.global.group_8 = val end,
            get = function(info) return self.db.global.group_8 end
          },
        }
      }
    }
  }

  AceConfig:RegisterOptionsTable("Dazzling Rod Tournament", options)
  self.optionsFrame = AceConfigDialog:AddToBlizOptions("Dazzling Rod Tournament", "Dazzling Rod Tournament")

  -- registering slash commands
  DazzlingRodTournament:RegisterChatCommand("drt", "ChatCmd")

  DazzlingRodTournament:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "COMBAT_LOG_EVENT_UNFILTERED")

end




function DazzlingRodTournament:OnEnable()
    -- Called when the addon is enabled
    -- DEBUG:DazzlingRodTournament:Print("Addon enabled.")
end

function DazzlingRodTournament:OnDisable()
    -- Called when the addon is disabled
    -- DEBUG:DazzlingRodTournament:Print("Addon disabled.")
end


